
set echo on

CREATE OR REPLACE PACKAGE error AS

   PROCEDURE log( p_owner         IN errors.package_owner%TYPE,
                  p_package       IN errors.package_name%TYPE,
                  p_procedure     IN errors.procedure_name%TYPE,
                  p_action        IN errors.action%TYPE,
                  p_business_date IN errors.business_date%TYPE,
                  p_business_key  IN errors.business_key%TYPE DEFAULT NULL,
                  p_error         IN VARCHAR2 DEFAULT NULL );

   bulk_exceptions EXCEPTION;
   PRAGMA EXCEPTION_INIT(bulk_exceptions, -24381);

   PROCEDURE log( p_owner         IN errors.package_owner%TYPE,
                  p_package       IN errors.package_name%TYPE,
                  p_procedure     IN errors.procedure_name%TYPE,
                  p_action        IN errors.action%TYPE,
                  p_business_date IN errors.business_date%TYPE,
                  p_business_data IN ANYDATA );

END;
/

CREATE OR REPLACE PACKAGE BODY error AS

   PROCEDURE log( p_owner         IN errors.package_owner%TYPE,
                  p_package       IN errors.package_name%TYPE,
                  p_procedure     IN errors.procedure_name%TYPE,
                  p_action        IN errors.action%TYPE,
                  p_business_date IN errors.business_date%TYPE,
                  p_business_key  IN errors.business_key%TYPE DEFAULT NULL,
                  p_error         IN VARCHAR2 DEFAULT NULL ) IS

      v_error errors.error_msg%TYPE := NVL(p_error,SQLERRM);

      PRAGMA AUTONOMOUS_TRANSACTION;

   BEGIN

      INSERT INTO errors 
         ( package_owner, package_name, procedure_name, action,
           business_date, business_key, error_ts, error_msg )
      VALUES
         ( p_owner, p_package, p_procedure, p_action,
           p_business_date, p_business_key, SYSTIMESTAMP, v_error );

      COMMIT;

   END log;

   PROCEDURE log( p_owner         IN errors.package_owner%TYPE,
                  p_package       IN errors.package_name%TYPE,
                  p_procedure     IN errors.procedure_name%TYPE,
                  p_action        IN errors.action%TYPE,
                  p_business_date IN errors.business_date%TYPE,
                  p_business_data IN ANYDATA ) IS

      v_plsql          VARCHAR2(1024);                 --<-- dynamic PL/SQL block
      v_type_name      VARCHAR2(61);                   --<-- underlying type name
      nt_error_indices number_ntt   := number_ntt();   --<-- offsets to bad data
      nt_business_keys varchar2_ntt := varchar2_ntt(); --<-- keys of bad data

   BEGIN

      /*
      || Determine the details of the collection type contained within the
      || ANYDATA parameter...
      */
      v_type_name := p_business_data.gettypename;

      /*
      || Determine where in our business data collection the bad records are...
      */
      FOR i IN 1 .. SQL%BULK_EXCEPTIONS.COUNT LOOP
         nt_error_indices.EXTEND;
         nt_error_indices(nt_error_indices.LAST) := SQL%BULK_EXCEPTIONS(i).error_index;
      END LOOP;

      /*
      || Build a PL/SQL block to accept the ANYDATA and a collection of error indices.
      || This will extract the collection from the ANYDATA instance and probe it
      || with the error indices from SQL%BULK_EXCEPTIONS, building up a return
      || collection of the bad data keys...
      */
      v_plsql := 'DECLARE '
              || '   n PLS_INTEGER; '
              || '   c ' || v_type_name || '; '
              || '   a SYS.ANYDATA  := :b1; '
              || '   x number_ntt   := :b2; '
              || '   o varchar2_ntt := varchar2_ntt(); '
              || 'BEGIN'
              || '   n := a.GetCollection(c); '
              || '   FOR i IN x.FIRST .. x.LAST LOOP '
              || '      o.EXTEND; '
              || '      o(o.LAST) := c(x(i)).print_key; '
              || '   END LOOP; '
              || '   :b3 := o; '
              || 'END; ';

      /*
      || Execute the PL/SQL string to return the bad data keys...
      */
      EXECUTE IMMEDIATE v_plsql USING IN  p_business_data,
                                      IN  nt_error_indices,
                                      OUT nt_business_keys;


      /*
      || Now we can log the errors...
      */
      FOR i IN 1 .. SQL%BULK_EXCEPTIONS.COUNT LOOP

         error.log( p_owner         => p_owner,
                    p_package       => p_package,
                    p_procedure     => p_procedure,
                    p_action        => p_action,
                    p_business_date => p_business_date,
                    p_business_key  => nt_business_keys(i),
                    p_error         => SQLERRM(-1*SQL%BULK_EXCEPTIONS(i).error_code) );
 
      END LOOP;

   END log;

END error;
/


